@@ -7,11 +7,12 @@ This module contains functionality for access to PolyAnalyst API.
 
 import contextlib
 import datetime
+import functools
 import os
 import pathlib
 import time
 import warnings
-from typing import Any, Dict, List, Tuple, Union, Optional, IO
+from typing import Any, Dict, List, Tuple, Union, Optional, IO, Iterator
 from urllib.parse import urljoin, urlparse
 
 import pytus
@@ -20,7 +21,7 @@ import requests
 import urllib3
 
 from . import __version__
-from .exceptions import *
+from .exceptions import APIException, ClientException, PAException, _WrapperNotFound
 
 __all__ = ['API', 'Project', 'PAException', 'ClientException', 'APIException']
 
@@ -28,7 +29,8 @@ __all__ = ['API', 'Project', 'PAException', 'ClientException', 'APIException']
 Response = Tuple[requests.Response, Any]
 Nodes = Dict[str, Dict[str, Union[int, str]]]
 Node = Dict[str, Union[str, int]]
-DataSet = List[Dict[str, Any]]
+_DataSet = List[Dict[str, Any]]
+JSON_VAL = Union[bool, str, int, float, None]
 
 urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
 
@@ -61,7 +63,7 @@ NodeTypes = [
 class API:
     """PolyAnalyst API
 
-    :param url: The scheme, host and port(if exists) of a PolyAnalyst server
+    :param url: The scheme, host and port(if exists) of a PolyAnalyst server \
         (e.g. ``https://localhost:5043/``, ``http://example.polyanalyst.com``)
     :param username: The username to login with
     :param password: (optional) The password for specified username
@@ -125,13 +127,13 @@ class API:
         """Returns api versions supported by PolyAnalyst server."""
         # the 'versions' endpoint was added in the 2191 polyanalyst's version
         try:
-            return self.request(urljoin(self.base_url, 'versions'), 'get')[1]
+            return self.request(urljoin(self.base_url, 'versions'), method='get')[1]
         except APIException:
             return ['1.0']
 
     def get_server_info(self) -> Optional[Dict[str, Union[int, str, Dict[str, str]]]]:
         """Returns general server information including build number, version and commit hashes."""
-        _, data = self.request(urljoin(self.url, 'server/info'), 'get')
+        _, data = self.request(urljoin(self.url, 'server/info'), method='get')
         return data
 
     def get_parameters(self) -> List[Dict[str, Union[str, List]]]:
@@ -142,19 +144,14 @@ class API:
         """Logs in to PolyAnalyst Server with user credentials."""
         credentials = {'uname': self.username, 'pwd': self.password}
         if self.ldap_server:
-            credentials['useLDAP'] = 1
+            credentials['useLDAP'] = '1'
             credentials['svr'] = self.ldap_server
 
-        resp, _ = self.request(
-            'login',
-            method='post',
-            params=credentials,
-        )
+        resp, _ = self.request('login', method='post', params=credentials)
         self.sid = resp.cookies['sid']
 
     def logout(self) -> None:
         """NOT IMPLEMENTED YET."""
-        pass
 
     def run_task(self, id: int) -> None:
         """Initiates scheduler task execution.
@@ -201,8 +198,8 @@ class API:
         kwargs['verify'] = self.certfile
         try:
             resp = self._s.request(method, url, **kwargs)
-        except requests.RequestException as e:
-            raise ClientException(e)
+        except requests.RequestException as exc:
+            raise ClientException(exc)
         else:
             return self._handle_response(resp)
 
@@ -219,6 +216,8 @@ class API:
         if isinstance(json, dict) and json.get('error'):
             with contextlib.suppress(KeyError):
                 error = json['error']
+                if 'The wrapper with the given GUID is not found on the server' == error['message']:
+                    raise _WrapperNotFound
                 error_msg = f"{error['title']}. Message: '{error['message']}'"
 
         # the old error response format handling
@@ -226,8 +225,7 @@ class API:
             if 'are not logged in' in response.text:
                 error_msg = 'You are not logged in to PolyAnalyst Server'
             elif 'operation is limited ' in response.text:
-                error_msg = ('Access to this operation is limited to project '
-                             'owners and administrator')
+                error_msg = 'Access to this operation is limited to project owners and administrator'
         elif response.status_code == 500:
             with contextlib.suppress(IndexError, TypeError):
                 if json[0] == 'Error':
@@ -235,8 +233,8 @@ class API:
         else:
             try:
                 response.raise_for_status()
-            except requests.HTTPError as e:
-                error_msg = e
+            except requests.HTTPError as exc:
+                error_msg = str(exc)
 
         with contextlib.suppress(NameError):
             raise APIException(error_msg, response.url, response.status_code)
@@ -259,8 +257,8 @@ class RemoteFileSystem:
         :param dest: (optional) path to the folder in the PolyAnalyst's user directory
         :param recursive: (optional) upload subdirectories recursively
 
-        :raise TypeError if ``source`` is not string or path-like object.
-               ValueError if ``source`` does not exists
+        :raise TypeError if ``source`` is not string or path-like object.\
+            ValueError if ``source`` does not exists
         """
         if not isinstance(source, (str, os.PathLike)):
             raise TypeError('The source parameter should be either string or path-like object.')
@@ -351,27 +349,23 @@ class RemoteFileSystem:
           ...     fs.upload_file(file, name='cars.csv', path='/data')
         """
         if file.tell():
-            warnings.warn("The file object's current position is not at the beginning of the file."
-                          "This will result in uploading only the part of the file!")
+            warnings.warn(
+                "The file object's current position is not at the beginning of the file."
+                "This will result in uploading only the part of the file!"
+            )
 
         file_name = name or os.path.basename(file.name)
         file_size = _get_file_size(file)
-        metadata = {'foldername': path}
 
         file_endpoint, _ = pytus.create(
             urljoin(self.api.url, 'file/upload'),
             file_name,
             file_size,
             session=self.api._s,
-            metadata=metadata
+            metadata={'foldername': path},
         )
 
-        pytus.resume(
-            file,
-            file_endpoint,
-            session=self.api._s,
-            offset=0
-        )
+        pytus.resume(file, file_endpoint, session=self.api._s, offset=0)
 
         # free up resources on the server if file is not uploaded completely
         try:
@@ -398,7 +392,10 @@ class Project:
         self._node_list: List[Node] = []
 
     def get_node_list(self) -> List[Node]:
-        """Returns a list of project nodes"""
+        """Returns a list of project nodes.
+
+        .. versionadded:: 0.15.0
+        """
         return self.api.get(
             'project/nodes',
             params={'prjUUID': self.uuid},
@@ -406,13 +403,13 @@ class Project:
         )['nodes']
 
     def get_execution_stats(self) -> List[Node]:
-        """Returns nodes execution statistics"""
-        return self.api.get(
-            'project/execution-statistics',
-            params={'prjUUID': self.uuid}
-        )['nodes']
+        """Returns nodes execution statistics.
 
-    def get_tasks(self) -> List[Dict[str, any]]:
+        .. versionadded:: 0.15.0
+        """
+        return self.api.get('project/execution-statistics', params={'prjUUID': self.uuid})['nodes']
+
+    def get_tasks(self) -> List[Dict[str, Any]]:
         """Returns task list info."""
         json = self.api.get('project/tasks', params={'prjUUID': self.uuid})
         # convert timestamp in milliseconds to python datetime
@@ -428,60 +425,68 @@ class Project:
         """Aborts the execution of all nodes in the project."""
         self.api.post('project/global-abort', json={'prjUUID': self.uuid})
 
-    def execute(self, *nodes: Union[str, Dict[str, str]], wait: bool = False) -> None:
+    def execute(self, *args: Union[str, Dict[str, str]], wait: bool = False) -> None:
         """Initiate execution of nodes and their children.
 
-        :param nodes: node names and/or dicts with name and type of nodes
+        :param args: node names and/or dicts with name and type of nodes
         :param wait: wait for nodes execution to complete
 
         Usage::
 
-        pass the names of the nodes to be executed separated by commas
           >>> prj.execute('Internet Source', 'Python')
 
         or, if there are several nodes in the project with the same name, pass
-        them as a dict with mandatory 'name' and 'type' keys (and because of this,
-        you can also pass items of prj.get_node_list())
+        them as a dict with mandatory ``name`` and ``type`` keys (and because of this,
+        you can also pass items of :func:`get_node_list`)
+
           >>> prj.execute(
-                  {'name': 'Example node', 'type': 'DataSource'},
-                  {'name': 'Example node', 'type': 'Dataset'},
-                  'Federated Search',
-                  prj.get_node_list()[1],
-              )
+          ...     {'name': 'Example node', 'type': 'DataSource'},
+          ...     {'name': 'Example node', 'type': 'Dataset'},
+          ...     'Federated Search',
+          ...     prj.get_node_list()[1],
+          ... )
+
+        use ``wait=True`` to wait for the passed nodes execution to complete.
+        Note that the other nodes execution started by this command will not be awaited.
 
-        use wait=True to wait for the passed nodes execution to complete. Note
-        that the other nodes execution started by this command will not be awaited.
           >>> prj.execute('Export to MS Word', wait=True)
         """
-        nodes = list(nodes)
-        for i, node in enumerate(nodes):
-            if isinstance(node, str):
-                nodes[i] = {'name': node, 'type': self._find_node(node)['type']}
+        nodes = []
+        for arg in args:
+            node = self._find_node(arg)
+            nodes.append({'name': node['name'], 'type': node['type']})
 
         self.api.post('project/execute', json={'prjUUID': self.uuid, 'nodes': nodes})
         if wait:
             for node in nodes:
-                self.wait_for_completion(node)
+                self.wait_for_completion(node)  # type: ignore
 
-    def preview(self, node: Union[str, Dict[str, str]]) -> DataSet:
+    def preview(self, node: Union[str, Dict[str, str]]) -> _DataSet:
         """Returns first 1000 rows of data from ``node``, texts and strings are
         cutoff after 250 symbols.
 
         :param node: node name or dict with name and type of node
 
+        .. deprecated:: 0.16.0
+           Use :method:: prj.dataset().preview() instead.
+
         Usage::
+
           >>> data_set = prj.preview('Python')
 
         or, pass dict if the project contains several nodes with the same name
           >>> data_set = prj.preview({'name': 'Python', 'type': 'Dataset'})
         """
-        if isinstance(node, str):
-            node = {'name': node, 'type': self._find_node(node)['type']}
+        return self.dataset(node).preview()
 
-        if node['type'] not in ('Dataset', 'DataSource'):
-            raise APIException(f"Invalid node type: {node['type']} (only 'Dataset' or 'DataSource' are available)")
+    def dataset(self, node: Union[str, Dict[str, str]]):
+        """Get dataset object
 
-        return self.api.get('dataset/preview', params={'prjUUID': self.uuid, **node})
+        :param node: node name or dict with name and type of the node
+
+        .. versionadded:: 0.16.0
+        """
+        return DataSet(self, self._find_node(node))
 
     def unload(self) -> None:
         """Unload the project from the memory and free system resources."""
@@ -502,18 +507,15 @@ class Project:
         This operation available only for project owner and administrators, and
         cannot be undone.
         """
-        self.api.post(
-            'project/delete',
-            json={'prjUUID': self.uuid, 'forceUnload': force_unload}
-        )
+        self.api.post('project/delete', json={'prjUUID': self.uuid, 'forceUnload': force_unload})
 
     def set_parameters(
-            self,
-            node: Union[str, Dict[str, str]],
-            node_type: str,
-            parameters: Dict[str, Any],
-            declare_unsync: bool = True,
-            hard_update: bool = True,
+        self,
+        node: Union[str, Dict[str, str]],
+        node_type: str,
+        parameters: Dict[str, Any],
+        declare_unsync: bool = True,
+        hard_update: bool = True,
     ) -> None:
         """Set default parameters of the selected Parameters node in the project.
 
@@ -521,53 +523,40 @@ class Project:
         :param node_type: node type, which parameters need to be set. The types are listed in NodeTypes.
         :param parameters: default parameters of the node to be set.
         :param declare_unsync: reset the status of the Parameters node.
-        :param hard_update: update every child node with new parameters if True, otherwise reset their statuses. Works
-        only if declare_unsync is True.
-
-        :raises ClientException when the `node` type is not Parameters
+        :param hard_update: update every child node with new parameters if True, \
+            otherwise reset their statuses. Works only if declare_unsync is True.
         """
-        if isinstance(node, str):
-            parameters_node = self._find_node(node)
-        else:
-            parameters_node = self._find_node(node['name'], node['type'])
-
-        if parameters_node['type'] != 'Parameters':
-            raise ClientException('The node type should be Parameters')
-
-        json: Optional[List[str]]
-        json = self.api.post(
+        warns: Optional[List[str]]
+        warns = self.api.post(
             'parameters/configure',
-            params={'prjUUID': self.uuid, 'obj': parameters_node['id']},
+            params={'prjUUID': self.uuid, 'obj': self._find_node(node)['id']},
             json={
                 'type': node_type,
                 'settings': parameters,
                 'declareUnsync': declare_unsync,
                 'hardUpdate': hard_update,
-            }
+            },
         )
-        if json:
-            for msg in json:
+        if warns:
+            for msg in warns:
                 warnings.warn(msg)
 
     def wait_for_completion(self, node: Union[str, Dict[str, str]]) -> bool:
-        """Waits for the node to complete the execution. Returns True if node have completed successfully and
-        False otherwise.
+        """Waits for the node to complete the execution. Returns True if node have
+        completed successfully and False otherwise.
 
         :param node: node name or dict with name and type of node
         """
-        if isinstance(node, str):
-            node = self._find_node(node)
-
         time.sleep(0.5)  # give pa time to update node statuses
         while True:
             self._update_node_list()
-            stats = self._find_node(node['name'], node['type'])
+            stats = self._find_node(node)
 
             if stats.get('errMsg'):
                 return False
             if stats['status'] == 'synchronized':
                 return True
-            elif stats['status'] == 'incomplete':
+            if stats['status'] == 'incomplete':
                 return False
 
             time.sleep(1)
@@ -575,25 +564,29 @@ class Project:
     def _update_node_list(self) -> None:
         self._node_list = self.get_node_list()
 
-    def _find_node(self, name: str, type_: Optional[str] = None) -> Optional[Node]:
+    def _find_node(self, node_: Union[str, Dict[str, str]]) -> Node:
+        if isinstance(node_, str):
+            name_, type_ = node_, None
+        else:
+            name_, type_ = node_['name'], node_['type']
+
         for node in self._node_list:
-            if type_:
-                if node['name'] == name and node['type'] == type_:
-                    return node
-            else:
-                if node['name'] == name:
-                    return node
+            if node['name'] == name_ and (type_ is None or node['type'] == type_):
+                return node
+
+        raise APIException(f"Node not found: name='{name_}', type='{type_}'", status_code=500)
 
     def get_nodes(self) -> Nodes:
         """Returns a dictionary of project's nodes information.
 
         The node value is a dict with a mandatory keys: id, type, status.
         It also may contain errMsg key if last node execution was failed.
+
+        .. deprecated:: 0.15.0
+           Use :func:`get_node_list` instead.
         """
         warnings.warn(
-            'Project.get_nodes() is deprecated, use Project.get_node_list() instead.',
-            DeprecationWarning,
-            stacklevel=2,
+            'Project.get_nodes() is deprecated, use Project.get_node_list() instead.', DeprecationWarning, stacklevel=2
         )
         json = self.api.get(
             'project/nodes',
@@ -607,15 +600,134 @@ class Project:
 
         Similar to :meth:`get_nodes() <Project.get_nodes>` but nodes contains
         extra information and the project statistics.
+
+        .. deprecated:: 0.15.0
+           Use :func:`get_execution_stats` instead.
         """
         warnings.warn(
             'Project.get_execution_statistics() is deprecated, use Project.get_execution_stats() instead.',
             DeprecationWarning,
             stacklevel=2,
         )
-        json = self.api.get(
-            'project/execution-statistics',
-            params={'prjUUID': self.uuid}
-        )
+        json = self.api.get('project/execution-statistics', params={'prjUUID': self.uuid})
         nodes = {node.pop('name'): node for node in json['nodes']}
         return nodes, json['nodesStatistics']
+
+
+def retry_on_invalid_guid(func):
+    @functools.wraps(func)
+    def wrapper(cls, *args, **kwargs):
+        try:
+            return func(cls, *args, **kwargs)
+        except _WrapperNotFound:
+            cls._update_guid()
+            return func(cls, *args, **kwargs)
+    return wrapper
+
+
+class DataSet:
+    def __init__(self, prj: Project, node: Node):
+        self._prj = prj
+        self._api = prj.api
+        self._node = node
+        # on purpose send wrong wrapperGuid(empty string) at first request to /dataset/* endpoints
+        # to create dataset wrapper on server and retrieve its' guid by @retry_on_invalid_guid
+        self.guid: str = ''
+
+    @retry_on_invalid_guid
+    def get_info(self) -> Dict[str, Any]:
+        return self._api.get('dataset/info', params={'wrapperGuid': self.guid})
+
+    @retry_on_invalid_guid
+    def get_progress(self) -> Dict[str, Union[int, str]]:
+        return self._api.get('dataset/progress', params={'wrapperGuid': self.guid})
+
+    def preview(self) -> _DataSet:
+        """Returns first 1000 rows of data from ``node``, texts and strings are
+        cutoff after 250 symbols.
+        """
+        return self._api.get(
+            'dataset/preview',
+            params={'prjUUID': self._prj.uuid, 'name': self._node['name'], 'type': self._node['type']},
+        )
+
+    def iter_rows(self, start: int = 0, stop: Optional[int] = None) -> Iterator[Dict[str, JSON_VAL]]:
+        """
+        Iterate over dataset rows.
+
+        :param start:
+        :param stop:
+
+        :raises ValueError if start or stop is out of datasets' row range
+
+        Usage::
+
+          download first 10 rows
+          >>> head = []
+          >>> for row in ds.iter_rows(0, 10):
+          ...     head.append(row)
+          download full dataset
+          >>> table = list(ds.iter_rows())
+          convert downloaded dataset to pandas.DataFrame
+          >>> df = pandas.DataFrame(table)
+        """
+        info = self.get_info()
+        max_row = info['rowCount']
+        if stop is None:
+            stop = max_row
+
+        # предпологается что если stop определен то пользователь в курсе количества строк в датасете
+        if not 0 <= start <= stop <= max_row:
+            raise ValueError(f'start and stop arguments must be within dataset row range: (0, {max_row})')
+
+        rows = self._values(stop)['table']
+        get_text = self._cell_text
+
+        class RowIterator:
+            def __init__(self):
+                self.idx = start
+
+            def __iter__(self):
+                return self
+
+            def __next__(self):
+                if self.idx >= stop:
+                    raise StopIteration
+
+                result = {}
+                for column in info['columnsInfo']:
+                    if column['flags'].get('getTextAlways'):
+                        result[column['title']] = get_text(self.idx, column['id'], column['title'])
+                    # elif column['type'] == 'DateTime':  # todo convert to python datetime?
+                    else:
+                        result[column['title']] = rows[self.idx][column['id']]
+
+                self.idx += 1
+                return result
+
+        return RowIterator()
+
+    def _update_guid(self) -> None:
+        self.guid = self._api.get(
+            'dataset/wrapper-guid',
+            params={'prjUUID': self._prj.uuid, 'obj': self._node['id']},
+        )['wrapperGuid']
+
+    @retry_on_invalid_guid
+    def _values(self, row_count: int) -> Dict[str, Union[List, Dict]]:
+        return self._api.get('dataset/values', json={'wrapperGuid': self.guid, 'rowCount': row_count})
+
+    @retry_on_invalid_guid
+    def _cell_text(self, row: int, col: int, _title) -> str:
+        return self._api.get(
+            'dataset/cell-text',
+            json={
+                'wrapperGuid': self.guid,
+                'row': row,
+                'col': col,
+                # todo remove next keys
+                'colTitle': _title,
+                'offset': 0,
+                'count': 0,
+            },
+        )['text']
